from aardvark import create_app, db
from aardvark.updater import AccountToUpdate
from flask import current_app
from swag_client.swag import get_all_accounts
import json
from bunch import Bunch
from flask_script import Manager, Command, Option
import better_exceptions # noqa
import Queue
import threading

manager = Manager(create_app)

ACCOUNT_QUEUE = Queue.Queue()
DB_LOCK = threading.Lock()
QUEUE_LOCK = threading.Lock()
UPDATE_DONE = False


class UpdateAccountThread(threading.Thread):
    global ACCOUNT_QUEUE, DB_LOCK, QUEUE_LOCK, UPDATE_DONE

    def __init__(self, thread_ID):
        self.thread_ID = thread_ID
        threading.Thread.__init__(self)
        self.app = current_app._get_current_object()

    def run(self):
        while not UPDATE_DONE:

            QUEUE_LOCK.acquire()

            if not ACCOUNT_QUEUE.empty():
                (account_num, role_name, arns) = ACCOUNT_QUEUE.get()

                self.app.logger.info("Thread #{} updating account {} with {} arns".format(
                                     self.thread_ID, account_num, 'all' if arns[0] == 'all' else len(arns)))

                QUEUE_LOCK.release()

                account = AccountToUpdate(self.app, account_num, role_name, arns)
                ret_code, aa_data = account.update_account()

                if ret_code != 0:  # retrieve wasn't successful, put back on queue
                    QUEUE_LOCK.acquire()
                    ACCOUNT_QUEUE.put((account_num, role_name, arns))
                    QUEUE_LOCK.release()

                self.app.logger.info("Thread #{} persisting data for account {}".format(self.thread_ID, account_num))

                DB_LOCK.acquire()
                persist_aa_data(self.app, aa_data)
                DB_LOCK.release()

            else:
                QUEUE_LOCK.release()


def persist_aa_data(app, aa_data):
    """
    Reads access advisor JSON file & persists to our database
    """
    from aardvark.model import AWSIAMObject, AdvisorData

    aa = json.loads(aa_data)

    with app.app_context():
        arn_cache = {}
        for arn, data in aa.items():
            if arn in arn_cache:
                item = arn_cache[arn]
            else:
                item = AWSIAMObject.get_or_create(arn)
                arn_cache[arn] = item
            for service in data:
                AdvisorData.create_or_update(item.id,
                                             service['lastAuthenticated'],
                                             service['serviceName'],
                                             service['serviceNamespace'],
                                             service['lastAuthenticatedEntity'],
                                             service['totalAuthenticatedEntities'])
        db.session.commit()


@manager.command
def drop_db():
    """ Drops the database. """
    db.drop_all()


@manager.command
def create_db():
    """ Creates the database. """
    db.create_all()


@manager.command
def config():
    """
    Creates a config file.

    SWAG_BUCKET = '...'
    SWAG_FILTER = '...'
    ROLENAME = '<ASSUME_ROLE_HERE>'
    REGION = 'us-east-1'
    NUM_THREADS = 5
    SQLALCHEMY_DATABASE_URI = 'postgresql://user:pass@localhost:5432/db' or
    SQLALCHEMY_DATABASE_URI = 'sqlite:///tmp/aardvark.db'
    PHANTOMJS = '/usr/local/bin/phantomjs'
    """
    bucket = raw_input('SWAG_BUCKET: ')
    swag_filter = raw_input('SWAG_FILTER:  ')
    role_name = raw_input('ROLENAME: ')
    region = raw_input('AWS REGION: ')
    db_uri = raw_input('DATABASE: ')
    num_threads = raw_input('Number of threads (5)') or 5  # testing shows problems with more than 6 threads
    phantom = raw_input('Path to phantomjs: ')

    log = """LOG_CFG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'filename': 'aardvark.log',
            'maxBytes': 10485760,
            'backupCount': 100,
            'encoding': 'utf8'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout'
        }
    },
    'loggers': {
        'aardvark': {
            'handlers': ['file', 'console'],
            'level': 'DEBUG'
        }
    }
}"""

    with open('config.py', 'w') as filedata:
        filedata.write('# Autogenerated config file\n')
        filedata.write('SWAG_BUCKET = "{bucket}"\n'.format(bucket=bucket))
        filedata.write('SWAG_FILTER = "{filter}"\n'.format(filter=swag_filter))
        filedata.write('ROLENAME = "{role}"\n'.format(role=role_name))
        filedata.write('REGION = "{region}"\n'.format(region=region))
        filedata.write('SQLALCHEMY_DATABASE_URI = "{uri}"\n'.format(uri=db_uri))
        filedata.write('PHANTOMJS = "{phantom}"\n'.format(phantom=phantom))
        filedata.write('SQLALCHEMY_TRACK_MODIFICATIONS = False\n')
        filedata.write('NUM_THREADS = "{num_threads}"\n'.format(num_threads=num_threads))
        filedata.write(log)


@manager.option('-a', '--accounts', dest='accounts', type=unicode, default='all')
@manager.option('-r', '--arns', dest='arns', type=unicode, default='all')
def update(accounts, arns):
    """
    Asks AWS for new Access Advisor information.
    """
    accounts = _prep_accounts(accounts)
    arns = arns.split(',')
    app = create_app()

    global ACCOUNT_QUEUE, QUEUE_LOCK, UPDATE_DONE

    role_name = app.config.get('ROLENAME')
    num_threads = app.config.get('NUM_THREADS') or 5

    if num_threads > 6:
        current_app.logger.warn('Greater than 6 threads seems to cause problems')

    QUEUE_LOCK.acquire()
    for account in accounts:
        ACCOUNT_QUEUE.put((account.metadata['account_number'], role_name, arns))
    QUEUE_LOCK.release()

    threads = []
    for thread_num in range(num_threads):
        thread = UpdateAccountThread(thread_num + 1)
        thread.start()
        threads.append(thread)

    while not ACCOUNT_QUEUE.empty():
        pass
    UPDATE_DONE = True


def _prep_accounts(account_names):
    """
    Convert CLI provided account names into list of accounts from SWAG.
    Considers account aliases as well as account names.
    Returns a set of Bunch'd accounts.
    """
    current_app.logger.info('getting bucket {}'.format(
                            current_app.config.get('SWAG_BUCKET')))

    service = {'services': {current_app.config.get('SWAG_FILTER') or
               'security_monkey': {'enabled': True}}, 'ours': True}

    accounts = get_all_accounts(bucket=current_app.config.get('SWAG_BUCKET'),
                                **service).get('accounts')

    if account_names == 'all':
        return [Bunch(account) for account in accounts]

    lookup = {account['name']: Bunch(account) for account in accounts}
    for account in accounts:
        for alias in account['alias']:
            lookup[alias] = Bunch(account)

    matching_accounts = list()
    account_names = account_names.split(',')
    account_names = {name.lower() for name in account_names}

    for name in account_names:
        if name not in lookup:
            current_app.logger.warn('Could not find an account named %s'
                                    % name)
            continue

        matching_accounts.append(lookup[name])

    return matching_accounts


class GunicornServer(Command):
    """
    This is the main GunicornServer server, it runs the flask app with gunicorn and
    uses any configuration options passed to it.
    You can pass all standard gunicorn flags to this command as if you were
    running gunicorn itself.
    For example:
    aardvark start_api -w 4 -b 127.0.0.0:8002
    Will start gunicorn with 4 workers bound to 127.0.0.0:8002
    """
    description = 'Run the app within Gunicorn'

    def get_options(self):
        options = []
        try:
            from gunicorn.config import make_settings
        except ImportError:
            # Gunicorn does not yet support Windows.
            # See issue #524. https://github.com/benoitc/gunicorn/issues/524
            # For dev on Windows, make this an optional import.
            print('Could not import gunicorn, skipping.')
            return options

        settings = make_settings()
        for setting, klass in settings.items():
            if klass.cli:
                if klass.action:
                    if klass.action == 'store_const':
                        options.append(Option(*klass.cli, const=klass.const, action=klass.action))
                    else:
                        options.append(Option(*klass.cli, action=klass.action))
                else:
                    options.append(Option(*klass.cli))
        return options

    def run(self, *args, **kwargs):
        from gunicorn.app.wsgiapp import WSGIApplication

        app = WSGIApplication()

        app.app_uri = 'aardvark:create_app()'
        return app.run()


def main():
    manager.add_command("start_api", GunicornServer())
    manager.run()


if __name__ == '__main__':
    manager.add_command("start_api", GunicornServer())
    manager.run()



from cloudaux.aws.iam import list_roles
from cloudaux.aws.sts import boto3_cached_conn
import requests
import json
import os
import tempfile
import urllib
import subprocess32
from subprocess32 import CalledProcessError

federation_base_url = 'https://signin.aws.amazon.com/federation'


class AccountToUpdate(object):
    def __init__(self, current_app, account_number, role_name, arns_list):
        self.current_app = current_app
        self.account_number = account_number
        self.role_name = role_name
        self.arn_list = arns_list
        self.conn_details = {
            'account_number': account_number,
            'assume_role': role_name,
            'session_name': 'aardvark',
            'region': self.current_app.config.get('REGION') or 'us-east-1'
        }

    def update_account(self):
        """
        Updates Access Advisor data for a given AWS account.
        1) Gets list of IAM Role ARNs in target account.
        2) Gets IAM credentials in target account.
        3) Exchanges IAM credentials for Signin Token.
        4) Calls PhantomJS to do the dirty work.
        5) Saves PhantomJS output to our DB.

        :return: Return code and JSON Access Advisor data for given account
        """
        arns = self._get_arns()

        if not arns:
            self.current_app.logger.warn("Zero ARNs collected. Exiting")
            exit(-1)

        creds = self._get_creds()
        token = _get_signin_token(creds)
        with tempfile.NamedTemporaryFile() as f:
            ret_code = self._call_phantom(token, list(arns), f.name)
            if ret_code == 0:
                return ret_code, f.read()
            else:
                return ret_code, None

    def _get_arns(self):
        """
        Gets a list of all Role ARNs in a given account, optionally limited by
        class property ARN filter
        :return: list of role ARNs
        """
        roles = list_roles(**self.conn_details)
        account_arns = set([role['Arn'] for role in roles])
        result_arns = set()
        for arn in self.arn_list:
            if arn.lower() == 'all':
                return account_arns

            if arn not in account_arns:
                self.current_app.logger.warn("Provided ARN {arn} not found in account.".format(arn=arn))
                continue

            result_arns.add(arn)

        return list(result_arns)

    def _get_creds(self):
        """
        Assumes into the target account and obtains Access Key, Secret Key, and Token

        :return: URL-encoded dictionary containing Access Key, Secret Key, and Token
        """
        client, credentials = boto3_cached_conn(
            'iam', account_number=self.account_number, assume_role=self.role_name, return_credentials=True)

        creds = json.dumps(dict(
            sessionId=credentials['AccessKeyId'],
            sessionKey=credentials['SecretAccessKey'],
            sessionToken=credentials['SessionToken']
        ))
        creds = urllib.quote(creds, safe='')
        return creds

    def _call_phantom(self, token, arns, output_file):
        """
        shells out to phantomjs.
        - Writes ARNs to a file that phantomjs will read as an input.
        - Phantomjs exchanges the token for session cookies.
        - Phantomjs then navigates to the IAM page and executes JavaScript
        to call GenerateServiceLastAccessedDetails for each ARN.
        - Every 10 seconds, Phantomjs calls GetServiceLastAccessedDetails
        - Phantom saves output to a file that is used by `persist()`

        :return: Exit code from phantomjs subprocess32
        """

        path = os.path.dirname(__file__)
        console_js = os.path.join(path, 'awsconsole.js')

        with tempfile.NamedTemporaryFile() as f:
            json.dump(arns, f)
            f.seek(0)
            try:
                p = subprocess32.Popen([
                    self.current_app.config.get('PHANTOMJS'),
                    console_js,
                    token,
                    f.name,
                    output_file],
                    stdout=subprocess32.PIPE, stderr=subprocess32.STDOUT)
                output, errs = p.communicate(timeout=1200)  # 20 mins
                self.current_app.logger.debug('Phantom Output: \n{}'.format(output))
                self.current_app.logger.debug('Phantom Errors: \n{}'.format(errs))
            except subprocess32.TimeoutExpired:
                self.current_app.logger.error('PhantomJS timed out')
                return 1  # return code 1 for timeout
            except CalledProcessError:
                self.current_app.logger.error('PhantomJS exited: {}'
                                              ''.format(p.returncode))
                return p.returncode
            else:
                self.current_app.logger.info('PhantomJS exited: 0')
                return 0


def _get_signin_token(creds):
    """
    Exchanges credentials dictionary for a signin token.

    1) Creates URL using credentials dictionary.
    2) Sends a GET request to that URL and parses the response looking for
    a signin token.

    :return: Signin Token
    """
    url = '{base}?Action=getSigninToken&Session={creds}'
    url = url.format(base=federation_base_url, creds=creds)
    return requests.get(url).json()['SigninToken']



from aardvark import db
import datetime
from flask import current_app
from sqlalchemy import BigInteger, Column, Integer, Text, TIMESTAMP
import sqlalchemy.exc
from aardvark.utils.sqla_regex import String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey


class AWSIAMObject(db.Model):
    """
    Meant to model AWS IAM Object Access Advisor.
    """
    __tablename__ = "aws_iam_object"
    id = Column(Integer, primary_key=True)
    arn = Column(String(2048), nullable=True, index=True, unique=True)
    lastUpdated = Column(TIMESTAMP)
    usage = relationship("AdvisorData", backref="item", cascade="all, delete, delete-orphan",
                         foreign_keys="AdvisorData.item_id")

    @staticmethod
    def get_or_create(arn):
        item = AWSIAMObject.query.filter(AWSIAMObject.arn == arn).scalar()

        added = False
        try:
            item = AWSIAMObject.query.filter(AWSIAMObject.arn == arn).scalar()
        except sqlalchemy.exc.SQLAlchemyException as e:
            current_app.logger.error('Database exception: {}'.format(e.message))

        if not item:
            item = AWSIAMObject(arn=arn, lastUpdated=datetime.datetime.utcnow())
            added = True
        else:
            item.lastUpdated = datetime.datetime.utcnow()
        db.session.add(item)

        # we only need a refresh if the object was created
        if added:
            db.session.commit()
            db.session.refresh(item)
        return item


class AdvisorData(db.Model):
    """
    Models certain IAM Access Advisor Data fields.

    {
      "totalAuthenticatedEntities": 1,
      "lastAuthenticatedEntity": "arn:aws:iam::XXXXXXXX:role/name",
      "serviceName": "Amazon Simple Systems Manager",
      "lastAuthenticated": 1489176000000,
      "serviceNamespace": "ssm"
    }
    """
    __tablename__ = "advisor_data"
    id = Column(Integer, primary_key=True)
    item_id = Column(Integer, ForeignKey("aws_iam_object.id"), nullable=False, index=True)
    lastAuthenticated = Column(BigInteger)
    serviceName = Column(String(128), index=True)
    serviceNamespace = Column(String(64), index=True)
    lastAuthenticatedEntity = Column(Text)
    totalAuthenticatedEntities = Column(Integer)

    @staticmethod
    def create_or_update(item_id, lastAuthenticated, serviceName, serviceNamespace, lastAuthenticatedEntity,
                         totalAuthenticatedEntities):
        try:
            item = AdvisorData.query.filter(AdvisorData.item_id == item_id).filter(AdvisorData.serviceNamespace ==
                                                                                   serviceNamespace).scalar()
        except sqlalchemy.exc.SQLAlchemyError as e:
            current_app.logger.error('Database error: {}'.format(e.message))

        if not item:
            item = AdvisorData(item_id=item_id,
                               lastAuthenticated=lastAuthenticated,
                               serviceName=serviceName,
                               serviceNamespace=serviceNamespace,
                               lastAuthenticatedEntity=lastAuthenticatedEntity,
                               totalAuthenticatedEntities=totalAuthenticatedEntities)
            db.session.add(item)
            return

        if lastAuthenticated > item.lastAuthenticated:
            item.lastAuthenticated = lastAuthenticated
            db.session.add(item)
